# Standard library imports
import numpy as np
import pandas as pd
import os
from matplotlib import pyplot as plt
from typing import Optional, Union

# Third-party package imports
from scipy.sparse import csc_matrix, csr_matrix, spmatrix
from scipy.sparse.csgraph import shortest_path
import networkx as nx
from IPython.display import display, HTML

from .utils import to_nparray, arrayable
from .path_finding import el_within_n_steps


def compute_flow_hitting_time(
    conn_df: Union[pd.DataFrame, spmatrix],
    flow_seed_idx: np.ndarray[int],
    flow_steps: int,
    flow_thre: float,
):
    """
    Compute hitting time for all cells in conn_df.
    Hitting time is the average number of hops required to reach a cell from a set of seed cells.
    The main algorithm is implemented in the 'navis' library (https://github.com/navis-org/navis).

    Args:
        conn_df (pd.DataFrame): DataFrame containing the connections with columns 'pre', 'post', and 'weight'.
        flow_seed_idx (np.ndarray): Array of seed cell indices.
        flow_steps (int): Number of steps for flow calculation.
        flow_thre (float): Threshold for activation in flow calculation.

    Returns:
        pd.DataFrame: DataFrame with columns 'idx' and 'hitting_time',
        where 'idx' is the cell index and 'hitting_time' is the computed hitting time.
    """

    try:
        from navis.models import BayesianTraversalModel, linear_activation_p
    except ImportError as e:
        raise ImportError(
            "The 'navis' library is required for computing information flow."
        ) from e

    # choose threshold for flow activation and define model
    def my_act(x):
        return linear_activation_p(x, min_w=0, max_w=flow_thre)

    if isinstance(conn_df, spmatrix):
        # convert sparse matrix to DataFrame
        coo = conn_df.tocoo()
        conn_df = pd.DataFrame({"pre": coo.row, "post": coo.col, "weight": coo.data})
    elif isinstance(conn_df, pd.DataFrame):
        # make sure the column names are correct
        if not all(col in conn_df.columns for col in ["pre", "post", "weight"]):
            raise ValueError(
                "Input DataFrame must contain columns 'pre', 'post', and 'weight'."
            )
    else:
        raise TypeError(
            "Input must be a sparse matrix or a DataFrame with columns 'pre', 'post', and 'weight'."
        )

    print(
        f"Computing hitting time for {len(set(coo.row)|set(coo.col))} cells... May take a while."
    )
    edges = conn_df[["pre", "post", "weight"]]
    edges.columns = ["source", "target", "weight"]
    model = BayesianTraversalModel(
        edges, flow_seed_idx, max_steps=flow_steps + 1, traversal_func=my_act
    )

    res = model.run()

    # compute hitting times from cmf (cumulative mass function)
    cmf_arr = np.stack(res["cmf"].values)
    hitting_prob = np.zeros(cmf_arr.shape)
    hitting_prob[:, 1:] = np.diff(cmf_arr, axis=1)
    hitting_time = np.dot(hitting_prob, np.arange(0, flow_steps + 1))

    # cmf of unreached and seed cell types is 0
    # change hitting time of unreached cell types to flow_steps
    idx_unreached = np.where(
        (hitting_time < 0.1) & (~np.isin(res["node"], flow_seed_idx))
    )[0]
    if len(idx_unreached) > 0:
        hitting_time[idx_unreached] = flow_steps

    flow_df = pd.DataFrame({"idx": res["node"].values, "hitting_time": hitting_time})

    return flow_df


def find_instance_flow(
    inprop: Union[spmatrix, pd.DataFrame],
    idx_to_group: dict,
    flow_seed_groups: list[str] = [
        "L1",
        "L2",
        "L3",
        "R7p",
        "R8p",
        "R7y",
        "R8y",
        "R7d",
        "R8d",
        "HBeyelet",
    ],
    file_path: str = None,
    save_flow: Optional[bool] = True,
    save_prefix: Optional[str] = "flow_",
    flow_steps: int = 20,
    flow_thre: float = 0.1,
) -> pd.DataFrame:
    """
    Get the hitting time for all cell groups. The hitting time is computed using the
    information flow algorithm (navis:https://github.com/navis-org/navis) for each
    neuron, and then taking the median across neurons of each cell group.

    Args:
        inprop (Union[spmatrix, pd.DataFrame]): Input sparse matrix or DataFrame
            representing connections. If a DataFrame, it should have columns 'pre',
            'post', and 'weight'.
        idx_to_group (dict): Dictionary mapping cell indices to their respective cell
            groups.
        flow_seed_groups (list): List of cell groups to be used as seeds for flow
            calculation.
        file_path (str): Path to the directory containing the hitting time data.
        save_flow (bool): Whether to save the computed hitting time to a CSV file.
            Defaults to True.
        save_prefix (str): Prefix for the saved file names.
        flow_seed_groups (list): List of cell group to be used as seeds for flow
            calculation.
        flow_steps (int): Number of steps for flow calculation.
        flow_thre (float): Threshold for flow calculation.

    Returns:
        pd.DataFrame: DataFrame with columns 'cell_group' and 'hitting_time', where
        'cell_group' is the name of the cell group and 'hitting_time' is the median
        hitting time for that group.
    """

    if file_path is None:
        # Check if running in Google Colab
        if "COLAB_GPU" in os.environ:
            # Running in Colab
            file_path = "/content/"
        else:
            # Running locally
            file_path = "./"

    # load hitting time or compute
    flow_file_name = os.path.join(
        file_path, f"{save_prefix}{flow_steps}step_{flow_thre}thre_hit.csv"
    )
    if os.path.exists(flow_file_name):
        flow_df = pd.read_csv(flow_file_name)
        if "cell_group" not in flow_df.columns:
            flow_df["cell_group"] = flow_df["idx"].map(idx_to_group)
    else:
        # if file_path doesn't exist, create it
        if not os.path.exists(file_path):
            os.makedirs(file_path)

        # get indices whose value is in flow_seed_groups
        flow_seed_idx = np.array(
            [
                idx
                for idx, cell_type in idx_to_group.items()
                if cell_type in flow_seed_groups
            ]
        )
        if len(flow_seed_idx) == 0:
            raise ValueError(
                f"No flow seed groups found in the provided idx_to_group mapping. "
                f"Please check the flow_seed_groups: {flow_seed_groups}."
            )
        flow_df = compute_flow_hitting_time(
            inprop, flow_seed_idx, flow_steps, flow_thre
        )
        # add cell group to flow_df
        # note: this only includes indices in flow_seed_idx, which could in theory be a
        # subset of cells in a cell type
        flow_df["cell_group"] = flow_df["idx"].map(idx_to_group)
        if save_flow:
            print(f"Saving hitting time to {flow_file_name}")
            flow_df.to_csv(flow_file_name, index=False)

    flow_type_df = (
        flow_df.groupby("cell_group")["hitting_time"]
        .median()
        .to_frame()
        .reset_index()
        .sort_values("hitting_time")
    )
    flow_type_file_name = os.path.join(
        file_path, f"{save_prefix}{flow_steps}step_{flow_thre}thre_hit_per_group.csv"
    )
    if not os.path.exists(flow_type_file_name):
        if save_flow:
            print(f"Saving flow group hitting time to {flow_type_file_name}")
            flow_type_df.to_csv(flow_type_file_name, index=False)

    return flow_type_df


def trim_inprop_by_flow(
    inprop: Union[spmatrix, pd.DataFrame],
    idx_to_group: dict,
    flow_seed_groups: list[str] = [
        "L1",
        "L2",
        "L3",
        "R7p",
        "R8p",
        "R7y",
        "R8y",
        "R7d",
        "R8d",
        "HBeyelet",
    ],
    file_path: str = None,
    save_flow: Optional[bool] = True,
    save_prefix: Optional[str] = "flow_",
    flow_steps: int = 20,
    flow_thre: float = 0.1,
    flow_diff_min: float = 0.5,
    flow_diff_max: float = 20,
) -> csc_matrix:
    """
    Trim connections based on hitting time assigned by information flow algorithm
    (navis: https://github.com/navis-org/navis). The hitting time is the mean
    number of hops required to reach a neuron from a neuron in flow_seed_groups.
    If the hitting time of the post neuron is larger than that of the pre neuron
    (i.e., the difference is between flow_diff_min and flow_diff_max), then the
    connection is interpreted as a feedforward connection and kept. For similar
    hitting times, the connection is interpreted as a lateral connection.
    If the hitting time of the pre neuron is larger than that of the post neuron,
    the connection is interpreted as a feedback connection.
    Lateral and feedback connections are removed.

    Args:
        inprop (csc_matrix): Input sparse matrix representing connections.
        idx_to_group (dict): Dictionary mapping cell indices to their respective cell groups.
        flow_seed_groups (list): List of cell groups to be used as seeds for flow calculation.
        file_path (str): Path to the directory containing the hitting time data.
        save_flow (bool): Whether to save the computed hitting time to a CSV file.
        flow_steps (int): Number of steps for flow calculation.
        flow_thre (float): Threshold for flow calculation.
        flow_diff_min (float): Minimum difference in hitting time for connection retention.
        flow_diff_max (float): Maximum difference in hitting time for connection retention.

    Returns:
        csc_matrix: sparse matrix for which pairs of connections have hitting time
        within specified range.
    """

    if file_path is None:
        # Check if running in Google Colab
        if "COLAB_GPU" in os.environ:
            # Running in Colab
            file_path = "/content/"
        else:
            # Running locally
            file_path = ""

    flow_type_df = find_instance_flow(
        inprop,
        idx_to_group,
        flow_seed_groups,
        file_path,
        save_flow,
        save_prefix,
        flow_steps,
        flow_thre,
    )  # has columns 'cell_group' and 'hitting_time'
    # note: this includes all neurons of e.g. a cell type
    meta = pd.DataFrame(idx_to_group.items(), columns=["idx", "cell_group"])
    meta = meta.merge(flow_type_df, how="inner", on="cell_group")

    if isinstance(inprop, pd.DataFrame):
        conn_df = inprop[["pre", "post", "weight"]]
    elif isinstance(inprop, spmatrix):
        # convert sparse matrix to DataFrame
        coo = inprop.tocoo()
        conn_df = pd.DataFrame({"pre": coo.row, "post": coo.col, "weight": coo.data})

    # add hitting time to conn_df
    conn_flow_df = (
        conn_df.merge(
            meta[["idx", "hitting_time"]],
            how="inner",
            left_on="post",
            right_on="idx",
        )
        .drop(columns=["idx"])
        .rename(columns={"hitting_time": "hitting_time_post"})
    )
    conn_flow_df = (
        conn_flow_df.merge(
            meta[["idx", "hitting_time"]],
            how="inner",
            left_on="pre",
            right_on="idx",
        )
        .drop(columns=["idx"])
        .rename(columns={"hitting_time": "hitting_time_pre"})
    )

    # remove connections for which difference in hitting time is too small or too large
    conn_flow_df = conn_flow_df[
        (
            conn_flow_df["hitting_time_post"] - conn_flow_df["hitting_time_pre"]
            >= flow_diff_min
        )
        & (
            conn_flow_df["hitting_time_post"] - conn_flow_df["hitting_time_pre"]
            <= flow_diff_max
        )
    ]
    inprop_flow = csc_matrix(
        (
            conn_flow_df["weight"],
            (conn_flow_df["pre"], conn_flow_df["post"]),
        ),
        shape=inprop.shape,
    )

    return inprop_flow


def find_shortest_paths(
    paths: pd.DataFrame, start_nodes: list[str], end_nodes: list[str]
) -> list[list[str]]:
    """
    Find the shortest paths between groups in start_nodes and end_nodes
    in a paths dataframe (paths is the output of find_path_iteratively).

    Args:
        paths (pd.DataFrame): DataFrame containing the path data, including
            columns 'weight', 'pre', and 'post'.
        start_nodes (list): List of 'pre' groups.
        end_nodes (list): List of 'post' groups.

    Returns:
        list: A list of shortest paths, where each path is a list of groups
            that connect the start and end nodes (ordered from start to end).
    """

    paths_unique = paths[["weight", "pre", "post"]].drop_duplicates()
    nodes_unique = np.unique(
        np.concatenate([paths_unique.pre.unique(), paths_unique.post.unique()])
    )
    pre = np.searchsorted(nodes_unique, paths_unique.pre.values)
    post = np.searchsorted(nodes_unique, paths_unique.post.values)
    graph_matrix = csr_matrix(
        (1 / paths_unique["weight"].values, (pre, post)),
        shape=(len(nodes_unique), len(nodes_unique)),
    )

    _, predecessors = shortest_path(
        csgraph=graph_matrix, directed=True, return_predecessors=True
    )

    def reconstruct_single_path(predecessors, start, end, node_names):
        """
        Helper function that reconstructs the path from start to end
        using the predecessors array.
        """
        idx_start = np.where(nodes_unique == start)[0]
        idx_end = np.where(nodes_unique == end)[0]
        if len(idx_start) == 0 or len(idx_end) == 0:
            return None
        path = [node_names[idx_end[0]]]
        i = idx_end[0]
        while i != idx_start[0]:
            i = predecessors[idx_start[0], i]
            if i == -9999:  # not reached
                return None
            path.append(node_names[i])
        return path[::-1]  # reverse

    shortest_paths = []
    for start_node in start_nodes:
        for end_node in end_nodes:
            if start_node != end_node:
                path = reconstruct_single_path(
                    predecessors, start_node, end_node, nodes_unique
                )
                if path is not None:
                    shortest_paths.append(path)

    return shortest_paths


def plot_flow_layered_paths(
    paths: pd.DataFrame,
    figsize: tuple = (10, 8),
    weight_decimals: int = 2,
    neuron_to_sign: dict | None = None,
    sign_color_map: dict = {1: "red", -1: "blue"},
    neuron_to_color: dict | None = None,
    edge_text: bool = True,
    node_text: bool = True,
    highlight_nodes: list[str] = [],
    interactive: bool = True,
    save_plot: bool = False,
    file_name: str = "layered_paths",
    label_pos: float = 0.7,
    default_neuron_color: str = "lightblue",
    default_edge_color: str = "lightgrey",
    node_size: int = 500,
) -> None:
    """
    Plots a directed graph of layered paths based on flow layers.
    Similar to the `plot_layered_paths` function, but the x-axis is defined by the
    flow layers of the nodes.

    Args:
        paths (pandas.DataFrame): A dataframe containing the columns 'pre', 'post',
            'weight', 'pre_layer', and 'post_layer'. Each row represents an edge in the
            graph. The 'pre' and 'post' columns refer to the source and target nodes,
            respectively, and 'weight' indicates the edge weight. 'pre_layer' and
            'post_layer' are the flow layers of the corresponding nodes.
        figsize (tuple, optional): A tuple indicating the size of the matplotlib figure.
            Defaults to (10, 8).
        weight_decimals (int, optional): The number of decimal places to display for
            edge weights. Defaults to 2.
        neuron_to_sign (dict, optional): A dictionary mapping neuron names (as they
            appear in path_df) to their signs (e.g. {'KCg-m': 1, 'Delta7': -1}).
            Can also use a dictionary to map neuron names to their neurotransmitter name.
            Defaults to None.
        sign_color_map (dict, optional): A dictionary used to color edges. Defaults is
            lightgrey but if `neuron_to_sign` is provided, the default is to color edges
            red if the pre-neuron is excitatory, and blue if inhibitory.
            If the `neuron_to_sign` values are neurotransmitter names, then provide
            a dictionary that maps neurotransmitter names to colors.
            If the keys of `sign_color_map` do not match the values of `neuron_to_sign`,
            a warning is printed and the default color is used for the difference.
        neuron_to_color (dict, optional): A dictionary mapping neuron names to colors.
            If not provided, a default color is used for all nodes. The keys should
            match the node names ('pre' and 'post') in paths. The difference is given
            the default color.
        edge_text (bool, optional): Whether to display edge weights as text on the plot.
            Defaults to True.
        node_text (bool, optional): Whether to display node names as text on the plot.
            Defaults to True.
        highlight_nodes (list[str], optional): A list of node names to highlight bold in
            the plot. Defaults to an empty list.
        interactive (bool, optional): Whether to create an interactive plot using pyvis.
            Defaults to False. If False, a static matplotlib plot is created.
        save_plot (bool, optional): Whether to save the plot to a file. Defaults to False.
        file_name (str, optional): The name of the file to save the plot. Defaults to
            "layered_paths" in the local directory (.html if interactive and .pdf if
            static).
        label_pos (float, optional): The position of the edge labels. Defaults to 0.7.
            Bigger values move the labels closer to the left of the edge.
            Only works if `interactive` is False.
        default_neuron_color (str, optional): The default color for nodes if no
            specific color is provided in `neuron_to_color`. Defaults to "lightblue".
        default_edge_color (str, optional): The default color for edges if no specific
            color is provided. Defaults to "lightgrey".
        node_size (int, optional): The size of the nodes in the plot. Defaults to 500.

    Returns:
        None: This function does not return a value. It generates a plot using
            matplotlib or pyvis.

    Note:
        The positions of the nodes are determined by a custom positioning function
            (`find_flow_positions`).
        This function requires the networkx library for graph operations and matplotlib
            for plotting. For interactive plots, it requires the pyvis library (where
            the node label has to be underneath the node).
    """

    def find_flow_positions(
        vertices: np.array, layers: np.array, height: float, width: float
    ) -> dict:
        """
        Find positions for the nodes depending on the layer flow.
        """
        # +0.5 at the end to make sure the max(layers) is included
        layer_bins = np.arange(-0.5 / 2, np.ceil(max(layers)) + 0.5, 1 / 2)
        layer_labels = layer_bins[1:] + layer_bins[0]
        layer_binned = pd.cut(layers, bins=layer_bins, labels=layer_labels)

        layer_w = width / max(layers)
        xy_pos = np.zeros((len(vertices), 2))
        for layer_b in np.unique(layer_binned):
            layer = layers[layer_binned == layer_b]
            layer_vertices = vertices[layer_binned == layer_b]
            if len(layer_vertices) > 0:
                # +1 so that when there aren't many vertices, they are in the middle
                layer_h = height / (len(layer_vertices) + 1)
                for index, v in enumerate(layer_vertices):
                    idx = np.where(vertices == v)[0][0]
                    xy_pos[idx, 1] = (index + 1) * layer_h
                    xy_pos[idx, 0] = layer[index] * layer_w - width / 2

        positions = dict(zip(vertices, xy_pos))

        return positions

    if paths.shape[0] == 0:
        raise ValueError("The provided DataFrame is empty.")

    path_df = paths.copy()

    # Rescale weights to be between 1 and 10
    weight_min = path_df.weight.min()
    weight_max = path_df.weight.max()
    if weight_max == weight_min:
        path_df["weight"] = 1
    else:
        path_df["weight"] = 1 + 9 * (path_df["weight"].values - weight_min) / (
            weight_max - weight_min
        )

    # Create the graph
    G = nx.from_pandas_edgelist(
        path_df,
        source="pre",
        target="post",
        edge_attr=True,
        create_using=nx.DiGraph(),
    )

    # Labels for nodes
    labels = {v: v for v in G.nodes()}

    # Generate positions
    layers = np.zeros(len(labels))
    for index, v in enumerate(labels):
        if v in path_df.pre.values:
            layers[index] = path_df[path_df.pre == v].pre_layer.values[0]
        elif v in path_df.post.values:
            layers[index] = path_df[path_df.post == v].post_layer.values[0]
    positions = find_flow_positions(
        np.array(list(labels.values())), layers, figsize[1] / 10, figsize[0] / 5
    )

    # Default color for nodes if not provided otherwise
    if neuron_to_color is None:
        neuron_to_color = {label: default_neuron_color for label in labels.values()}
    else:
        # Ensure neuron_to_color has all labels, even if not in the dictionary
        nodediff = set(labels.values()) - set(neuron_to_color.keys())
        if len(nodediff) > 0:
            neuron_to_color.update(dict.fromkeys(nodediff, default_neuron_color))

    node_colors = []
    for v in G.nodes():
        node_colors.append(neuron_to_color[labels[v]])

    if neuron_to_sign is not None:
        signdiff = set(neuron_to_sign.values()) - set(sign_color_map.keys())
        if len(signdiff) > 0:
            print(
                "Warning: Some values in neuron_to_sign are not in the keys of sign_color_map. Using default color for those edges."
            )
            # this is taken care of below with sign_color_map.get(pre_neuron, default_edge_color)

    # Specify edge colour based on pre-neuron sign, if available
    edge_colors = []
    for u, _ in G.edges():
        pre_neuron = labels[u]
        if neuron_to_sign and pre_neuron in neuron_to_sign:
            edge_colors.append(
                sign_color_map.get(neuron_to_sign[pre_neuron], default_edge_color)
            )
        else:
            edge_colors.append(default_edge_color)

    if interactive:
        try:
            from pyvis.network import Network
        except ImportError as e:
            raise ImportError(
                "Please install pyvis for interactive plots: pip install pyvis"
            ) from e

        # create pyvis graph
        net2 = Network(
            directed=True, layout=False, notebook=True, cdn_resources="in_line"
        )
        canvas_h_px = int(figsize[1] * 80)  # tweak the multiplier if needed
        net2.height = f"{canvas_h_px}px"  # explicit px keeps Colab happy
        net2.width = "100%"  # stretch side-to-side
        net2.from_nx(G)

        node_colors_dict = dict(zip(G.nodes(), node_colors))
        edge_colors_dict = dict(zip(G.edges(), edge_colors))
        for v in net2.nodes:
            v["x"], v["y"] = positions.get(v["id"])
            v["x"] = v["x"] * int(figsize[0] * 80)
            v["y"] = v["y"] * canvas_h_px
            if node_text:
                v["label"] = labels[v["id"]]
            else:
                v["label"] = ""
            v["color"] = node_colors_dict[v["id"]]
            v["size"] = node_size / 50
            v["font"] = {"size": 24}
            if labels[v["id"]] in highlight_nodes:
                v["font"] = {"face": "arial black"}
            else:
                v["font"] = {"face": "arial"}

        for edge in net2.edges:
            u, v = edge["from"], edge["to"]
            edge["color"] = edge_colors_dict[(u, v)]
            if edge_text:
                edge["label"] = (
                    f"{(weight_min+(weight_max - weight_min)*(edge['width']-1)/9):.{weight_decimals}f}"
                )
                edge["font"] = {"size": 18, "face": "arial"}

        # Set physics options for the network with high spring constant
        # to keep straighter arrows when nodes are moved around
        net2.set_options(
            """
        var options = {
        "physics": {
            "enabled": true,
            "solver": "forceAtlas2Based",
            "forceAtlas2Based": {
            "springConstant": 0.1
            },
            "minVelocity": 0.1
        },
        "nodes": {
            "physics": false
        }
        }
        """
        )
        if "COLAB_GPU" in os.environ:
            # display of plot in ipynb only works in colab
            html = net2.generate_html(notebook=True)
            if save_plot:
                with open(f"{file_name}.html", "w") as f:
                    f.write(html)

            display(HTML(html))
        else:
            # if running locally, will just open in browser
            if save_plot:
                net2.write_html(str(file_name) + ".html")
                print(f"Interactive graph saved as {file_name}.html")

            net2.show(str(file_name) + ".html", notebook=False)

    else:

        fig, ax = plt.subplots(figsize=figsize)
        nx.draw(
            G,
            pos=positions,
            with_labels=False,
            node_size=node_size,
            node_color=node_colors,
            arrows=True,
            arrowstyle="-|>",
            arrowsize=10,
            width=[G[u][v]["weight"] for u, v in G.edges()],
            edge_color=edge_colors,
            ax=ax,
        )

        if node_text:
            # label highlighted and normal nodes separately
            bold_nodes = [n for n in G.nodes() if labels[n] in highlight_nodes]
            normal_nodes = [n for n in G.nodes() if labels[n] not in highlight_nodes]

            nx.draw_networkx_labels(
                G,
                pos=positions,
                labels={n: labels[n] for n in normal_nodes},
                font_family="sans-serif",
                font_weight="normal",
                font_size=14,
                ax=ax,
            )
            nx.draw_networkx_labels(
                G,
                pos=positions,
                labels={n: labels[n] for n in bold_nodes},
                font_family="sans-serif",
                font_weight="bold",
                font_size=14,
                ax=ax,
            )

        if edge_text:
            edge_labels = {
                (
                    u,
                    v,
                    # transform back to the original weight values
                ): f"{(weight_min+(weight_max - weight_min)*(G[u][v]['weight']-1)/9):.{weight_decimals}f}"
                for u, v in G.edges()
            }
            nx.draw_networkx_edge_labels(
                G,
                pos=positions,
                edge_labels=edge_labels,
                label_pos=label_pos,
                font_size=14,
                rotate=False,
                ax=ax,
            )

        ax.set_ylim(0, 1)
        if save_plot:
            fig.savefig(file_name + ".pdf")
            print(f"Graph saved as {file_name}.pdf")
        plt.show()


def layered_el(
    inprop: spmatrix,
    steps: list,
    inidx: arrayable,
    outidx: arrayable,
    n: int,
    idx_to_group: dict,
    threshold: float = 0,
    flow_steps: int = 20,
    flow_thre: float = 0.1,
    flow: pd.DataFrame | None = None,
):
    """
    First finds paths within `n` steps, given the `threshold` (applied to direct
    connections), grouped by `idx_to_group`, and adds flow layers to the edge list based
    on the information flow hitting time (navis: https://github.com/navis-org/navis).

    Args:
        inprop (spmatrix): Input sparse matrix representing connections.
        steps (list): A list of connectivity matrices, e.g. the result from
            `compress_paths()`.
        inidx (np.ndarray): Array of input indices.
        outidx (np.ndarray): Array of output indices.
        n (int): The maximum number of hops. n=1 for direct connections.
        idx_to_group (dict): Dictionary mapping cell indices to their respective cell
            groups.
        threshold (float): The threshold for the weight of the direct
            connection between pre and post, after grouping by `idx_to_group`. Defaults
            to 0.
        flow_steps (int): Number of steps for flow calculation. Defaults to 20.
        flow_thre (float): Threshold for flow calculation. Defaults to 0.1.
        flow (pd.DataFrame, optional): DataFrame containing the flow hitting time.
            If provided, it should have columns 'cell_group' and 'hitting_time'.
            If None, the flow hitting time is computed from `inprop` and `idx_to_group`.

    Returns:
        pd.DataFrame: DataFrame containing the grouped edge list with flow layers.
    """

    inidx = to_nparray(inidx)
    outidx = to_nparray(outidx)

    # get edge list, both grouped by idx_to_group, and with raw indices
    grouped, rawel = el_within_n_steps(
        inprop,
        steps,
        inidx,
        outidx,
        n,
        threshold,
        idx_to_group,
        idx_to_group,
        return_raw_el=True,
    )
    rawel["pre_type"] = rawel.pre.map(idx_to_group)
    rawel["post_type"] = rawel.post.map(idx_to_group)
    # only keep the neurons in the types that pass the threshold
    rawel = rawel[rawel.pre_type.isin(grouped.pre) & rawel.post_type.isin(grouped.post)]

    # information flow
    if flow is None:
        # compute flow hitting time
        flow = find_instance_flow(
            rawel,
            idx_to_group,
            set([grp for idx, grp in idx_to_group.items() if idx in inidx]),
            save_flow=False,
            flow_steps=flow_steps,
            flow_thre=flow_thre,
        )
    else:
        # use provided flow hitting time
        if not isinstance(flow, pd.DataFrame):
            raise TypeError("Flow must be a pandas DataFrame.")
        if "cell_group" not in flow.columns or "hitting_time" not in flow.columns:
            raise ValueError(
                "Flow DataFrame must contain 'cell_group' and 'hitting_time' columns."
            )
    type_layer = dict(zip(flow.cell_group, flow.hitting_time))

    grouped["pre_layer"] = grouped.pre.map(type_layer)
    grouped["post_layer"] = grouped.post.map(type_layer)
    grouped.fillna(0, inplace=True)
    return grouped



import io
import pkgutil
import os
from typing import Optional, Union

import numpy as np
import pandas as pd
import plotly.graph_objects as go
import matplotlib.pyplot as plt

DATA_SOURCES: dict[str, str] = {
    "DoOR_adult": "data/DoOR/processed_door_adult.csv",
    "DoOR_adult_sfr_subtracted": "data/DoOR/processed_door_adult_sfr_subtracted.csv",
    "Dweck_adult_chem": "data/Dweck2018/adult_chem2glom.csv",
    "Dweck_adult_fruit": "data/Dweck2018/adult_fruit2glom.csv",
    "Dweck_larva_chem": "data/Dweck2018/larva_chem2or.csv",
    "Dweck_larva_fruit": "data/Dweck2018/larva_fruit2or.csv",
    "Nern2024": "data/Nern2024/ME-columnar-cells-hex-location.csv",
    "Matsliah2024": "data/Matsliah2024/fafb_right_vis_cols.csv",
    "Badel2016_PN": "data/Badel2016/Badel2016.csv",
    "Zhao2024": "data/Zhao2024/ucl_hex_right_20240701_tomale.csv",
}


def load_dataset(dataset: str) -> pd.DataFrame:
    """
    Load the dataset from the package data folder. These datasets have been
    preprocessed to work with connectomics data. The preprocessing scripts are
    in this repository: https://github.com/YijieYin/interpret_connectome.

    Args:
        dataset: (str) The name of the dataset to load. Options are:

            - 'DoOR_adult': mapping from glomeruli to chemicals, from Munch and Galizia DoOR dataset (https://www.nature.com/articles/srep21841).
            - 'DoOR_adult_sfr_subtracted': mapping from glomeruli to chemicals, with spontaneous firing rate subtracted. There are therefore negative values.
            - 'Dweck_adult_chem': mapping from glomeruli to chemicals extracted from fruits, from Dweck et al. 2018 (https://www.cell.com/cell-reports/abstract/S2211-1247(18)30663-6). Firing rates normalised to between 0 and 1.
            - 'Dweck_adult_fruit': mapping from glomeruli to fruits, from Dweck et al. 2018. Number of responses normalised to between 0 and 1.
            - 'Dweck_larva_chem': mapping from olfactory receptors to chemicals, from Dweck et al. 2018. Firing rates normalised to between 0 and 1.
            - 'Dweck_larva_fruit': mapping from olfactory receptors to fruits from Dweck et al. 2018. Number of responses normalised to between 0 and 1.
            - 'Nern2024': columnar coordinates of individual cells from a collection of columnar cell types within the medulla of the right optic lobe, from Nern et al. 2024 (https://www.biorxiv.org/content/10.1101/2024.04.16.589741v2).
            - 'Matsliah2024': columnar coordinates of individual cells from a collection of columnar cell types in the right optic lobe from FAFB, from Matsliah et al. 2024 (https://www.nature.com/articles/s41586-024-07981-1).
            - 'Badel2016_PN': mapping from olfactory projection neurons to odours, from Badel et al. 2016 (https://www.cell.com/neuron/fulltext/S0896-6273(16)30201-X).
            - 'Zhao2024': mapping from hexagonal coordinates to 3D coordinates, update from Zhao et al. 2022 (https://www.biorxiv.org/content/10.1101/2022.12.14.520178v1).

    Returns:
        pd.DataFrame: The dataset as a pandas DataFrame. For the adult, the glomeruli
        are in the rows. For the larva, receptors are in the rows.
    """

    try:
        data = pkgutil.get_data("connectome_interpreter", DATA_SOURCES[dataset])
    except KeyError as exc:
        raise ValueError(
            "Dataset not recognized. Please choose from {}".format(
                list(DATA_SOURCES.keys())
            )
        ) from exc

    return pd.read_csv(io.BytesIO(data), index_col=0)


def map_to_experiment(df, dataset=None, custom_experiment=None):
    """
    Map the connectomics data to experimental data. For example, if odour1
    excites neuron1 0.5, and neuron2 0.6; both neuron1 and neuron2 output to
    neuron3 (0.7 and 0.8 respectively), then the output of neuron3 to odour1
    is 0.5*0.7 + 0.6*0.8 = 0.83. The result would only be 1 if a stimulus
    excites neurons 100%, and those neurons constitue 100% of the downstream
    neuron's input.

    Args:
        df (pd.DataFrame): The connectivity data. Standardised input (e.g. glomeruli,
            receptors) in rows, observations (target neurons) in columns.
        dataset (str): The name of the dataset to load. Options are:

            - 'DoOR_adult': mapping from glomeruli to chemicals, from Munch and Galizia DoOR dataset (https://www.nature.com/articles/srep21841).
            - 'DoOR_adult_sfr_subtracted': mapping from glomeruli to chemicals, with spontaneous firing rate subtracted. There are therefore negative values.
            - 'Dweck_adult_chem': mapping from glomeruli to chemicals extracted from fruits, from Dweck et al. 2018 (https://www.cell.com/cell-reports/abstract/S2211-1247(18)30663-6). Firing rates normalised to between 0 and 1.
            - 'Dweck_adult_fruit': mapping from glomeruli to fruits, from Dweck et al. 2018. Number of responses normalised to between 0 and 1.
            - 'Dweck_larva_chem': mapping from olfactory receptors to chemicals, from Dweck et al. 2018. Firing rates normalised to between 0 and 1.
            - 'Dweck_larva_fruit': mapping from olfactory receptors to fruits, from Dweck et al. 2018. Number of responses normalised to between 0 and 1.
            - 'Nern2024': columnar coordinates of individual cells from a collection of columnar cell types within the medulla of the right optic lobe, from Nern et al. 2024.
            - 'Badel2016_PN': mapping from olfactory projection neurons to odours, from Badel et al. 2016 (https://www.cell.com/neuron/fulltext/S0896-6273(16)30201-X).

        custom_experiment (pd.DataFrame): A custom experimental dataset to compare the
            connectomics data to. The row indices of this dataframe must match the row
            indices of df. They are the units of comparison (e.g. glomeruli).

    Returns:
        pd.DataFrame: The similarity between the connectomics data and the experimental
        data. Rows are neurons, columns are external stimulus.
    """

    # try:
    #     from sklearn.metrics.pairwise import cosine_similarity
    # except ImportError as e:
    #     raise ImportError(
    #         "To use this function, please install scikit-learn. You can
    # install it with 'pip install scikit-learn'.") from e
    if dataset is not None and custom_experiment is not None:
        raise ValueError(
            "Please provide either a dataset or a custom_experiment, not both."
        )
    if dataset is None and custom_experiment is None:
        raise ValueError("Please provide either a dataset or a custom_experiment.")
    if dataset is not None:
        data = load_dataset(dataset)
    else:
        data = custom_experiment

    # take the intersection of glomeruli
    data = data[data.index.isin(df.index)]
    df_intersect = df[df.index.isin(data.index)]
    df_intersect = df_intersect.reindex(data.index)

    # multiply the correpsonding values using matmul
    target2chem = np.dot(df_intersect.values.T, data.values)
    # Assign appropriate column names
    target2chem = pd.DataFrame(
        target2chem, index=df_intersect.columns, columns=data.columns
    )
    return target2chem


def hex_heatmap(
    df: pd.Series | pd.DataFrame,
    style: Optional[dict] = None,
    sizing: Optional[dict] = None,
    dpi: int = 72,
    custom_colorscale: Optional[Union[list, str]] = None,
    global_min: Optional[float] = None,
    global_max: Optional[float] = None,
    dataset: Optional[str] = "mcns_right",
    value_name: str = "weight",
) -> go.Figure:
    """
    Generate a hexagonal heat map plot of the data. The index of the data
    should be formatted as strings of the form '-12,34', where the first
    number is the x-coordinate and the second number is the y-coordinate.

    Args:
        df (pd.Series | pd.DataFrame): The data to plot. If a Series, it will be plotted
            as a single trace. If a DataFrame, each column will generate a separate
            frame in the plot.
        style (Optional[dict]): Dict containing styling formatting variables. Possible
            keys are:

                - 'font_type': str, default='arial'
                - 'linecolor': str, default='black'
                - 'papercolor': str, default='rgba(255,255,255,255)' (white)

        sizing (Optional[dict]): Dict containing size formatting variables. Possible
            keys are:

                - 'fig_width': int, default=260 (mm)
                - 'fig_height': int, default=220 (mm)
                - 'fig_margin': int, default=0 (mm)
                - 'fsize_ticks_pt': int, default=20 (points)
                - 'fsize_title_pt': int, default=20 (points)
                - 'markersize': int, default=18 if dataset='mcns_right', 20 if dataset='fafb_right'
                - 'ticklen': int, default=15
                - 'tickwidth': int, default=5
                - 'axislinewidth': int, default=3
                - 'markerlinewidth': int, default=0.9
                - 'cbar_thickness': int, default=20
                - 'cbar_len': float, default=0.75

        dpi (int): Dots per inch for the output figure. Standard is 72 for screen/SVG/PDF.
            Use higher values (e.g., 300) for print-quality output.
        custom_colorscale (Optional[Union[list, str]]): Custom colorscale for the
            heatmap. If None, defaults to white-to-blue colorscale [[0, "rgb(255, 255,
            255)"], [1, "rgb(0, 20, 200)"]].
        global_min (Optional[float]): Global minimum value for the color scale. If None,
            the minimum value of the data is used but if that is negative, use 0.
        global_max (Optional[float]): Global maximum value for the color scale. If None,
            the maximum value of the data is used.
        dataset (str): Default='mcns_right'. The dataset to use for the hexagon
            locations. Options are:

                - 'mcns_right': columnar coordinates of individual cells from columnar cell types: L1, L2, L3, L5, Mi1, Mi4, Mi9, C2, C3, Tm1, Tm2, Tm4, Tm9, Tm20, T1, within the medulla of the right optic lobe, from Nern et al. 2024.
                - 'fafb_right': columnar coordinates of individual cells from columnar cell types, in the right optic lobe of FAFB, from Matsliah et al. 2024.

    Returns:
        fig : go.Figure
    """

    def bg_hex():
        """
        Generate a scatter plot of the background hexagons."
        """
        goscatter = go.Scatter(
            x=background_hex["x"],
            y=background_hex["y"],
            mode="markers",
            marker_symbol=symbol_number,
            marker={
                "size": sizing["markersize"],
                "color": "white",
                "line": {
                    "width": sizing["markerlinewidth"],
                    "color": "lightgrey",
                },
            },
            showlegend=False,
        )
        return goscatter

    def data_hex(aseries):
        """
        Generate a scatter plot of the data hexagons.""
        """
        goscatter = go.Scatter(
            x=x_vals,
            y=y_vals,
            mode="markers",
            marker_symbol=symbol_number,
            customdata=np.stack([x_vals, y_vals, aseries.values], axis=-1),
            hovertemplate="x: %{customdata[0]}<br>y: %{customdata[1]}<br>%{text}: %{customdata[2]:.4f}",
            text=[value_name] * len(aseries),
            marker={
                "cmin": global_min,
                "cmax": global_max,
                "size": sizing["markersize"],
                "color": aseries.values,
                "line": {
                    "width": sizing["markerlinewidth"],
                    "color": "lightgrey",
                },
                "colorbar": {
                    "orientation": "v",
                    "outlinecolor": style["linecolor"],
                    "outlinewidth": sizing["axislinewidth"],
                    "thickness": sizing["cbar_thickness"],
                    "len": sizing["cbar_len"],
                    "tickmode": "array",
                    "ticklen": sizing["ticklen"],
                    "tickwidth": sizing["tickwidth"],
                    "tickcolor": style["linecolor"],
                    "tickfont": {
                        "size": fsize_ticks_px,
                        "family": style["font_type"],
                        "color": style["linecolor"],
                    },
                    "tickformat": ".5f",
                    "title": {
                        "font": {
                            "family": style["font_type"],
                            "size": fsize_title_px,
                            "color": style["linecolor"],
                        },
                        "side": "right",
                    },
                },
                "colorscale": custom_colorscale,
            },
            showlegend=False,
        )
        return goscatter

    # Default styling and sizing parameters to use if not specified.
    default_style = {
        "font_type": "arial",
        "markerlinecolor": "rgba(0,0,0,0)",  # transparent
        "linecolor": "black",
        "papercolor": "rgba(255,255,255,255)",
    }

    if dataset == "mcns_right":
        markersize = 18
    elif dataset == "fafb_right":
        markersize = 20
    else:
        # raise error
        raise ValueError(
            "Dataset not recognized. Currently available datasets are 'mcns_right', "
            "'fafb_right'."
        )

    default_sizing = {
        "fig_width": 260,  # units = mm
        "fig_height": 220,  # units = mm
        "fig_margin": 0,
        "fsize_ticks_pt": 20,
        "fsize_title_pt": 20,
        "markersize": markersize,
        "ticklen": 15,
        "tickwidth": 5,
        "axislinewidth": 3,
        "markerlinewidth": 0.5,  # 0.9,
        "cbar_thickness": 20,
        "cbar_len": 0.75,
    }

    # If style is provided, update default_style with user values
    if style is not None:
        default_style.update(style)
    style = default_style

    if sizing is not None:
        default_sizing.update(sizing)
    sizing = default_sizing

    # Constants for unit conversion
    POINTS_PER_INCH = 72  # Typography standard: 1 point = 1/72 inch
    MM_PER_INCH = 25.4  # Standard conversion: 1 inch = 25.4 mm

    # sizing of the figure and font
    pixelsperinch = dpi  # Use the provided DPI value
    pixelspermm = pixelsperinch / MM_PER_INCH

    # Default colorscale
    if custom_colorscale is None:
        custom_colorscale = [[0, "rgb(255, 255, 255)"], [1, "rgb(0, 20, 200)"]]

    area_width = (sizing["fig_width"] - sizing["fig_margin"]) * pixelspermm
    area_height = (sizing["fig_height"] - sizing["fig_margin"]) * pixelspermm

    fsize_ticks_px = sizing["fsize_ticks_pt"] * (1 / POINTS_PER_INCH) * pixelsperinch
    fsize_title_px = sizing["fsize_title_pt"] * (1 / POINTS_PER_INCH) * pixelsperinch

    # Get global min and max for consistent color scale
    # minimum of 0 and df.values.min()
    vals = df.to_numpy()
    if global_min is None:
        global_min = min(0, vals.min())
    if global_max is None:
        global_max = vals.max()

    # Symbol number to choose to plot hexagons
    symbol_number = 15

    # load all hex coordinates
    if dataset == "mcns_right":
        background_hex = load_dataset("Nern2024")
    elif dataset == "fafb_right":
        background_hex = load_dataset("Matsliah2024")
    else:
        # raise error
        raise ValueError(
            "Dataset not recognized. Currently available datasets are 'mcns_right', "
            "'fafb_right'."
        )
    # only get the unique combination of 'x' and 'y' columns
    background_hex = background_hex.drop_duplicates(subset=["x", "y"])

    # initiate plot
    fig = go.Figure()
    fig.update_layout(
        autosize=False,
        height=area_height,
        width=area_width,
        margin={"l": 0, "r": 0, "b": 0, "t": 0, "pad": 0},
        paper_bgcolor=style["papercolor"],
        plot_bgcolor=style["papercolor"],
    )
    fig.update_xaxes(
        showgrid=False, showticklabels=False, showline=False, visible=False
    )
    fig.update_yaxes(
        showgrid=False, showticklabels=False, showline=False, visible=False
    )

    # Convert index values (formatted as '-12,34') into separate x and y coordinates
    df = df[(df.index != "nan") & (~df.index.isnull())]
    coords = [tuple(map(float, idx.split(","))) for idx in df.index]
    x_vals, y_vals = zip(*coords)  # Separate into x and y lists

    if isinstance(df, pd.Series) or len(df.columns) == 1:
        if isinstance(df, pd.DataFrame):
            df = df.iloc[:, 0]
        fig.add_trace(bg_hex())
        fig.add_trace(data_hex(df))

    elif isinstance(df, pd.DataFrame):
        # Adjust figure size - add extra height for slider
        slider_height = 100  # pixels
        area_height += slider_height

        # Create frames for slider
        frames = []
        slider_steps = []

        # Add base layout
        fig.update_layout(
            autosize=False,
            height=area_height,
            width=area_width,
            margin={
                "l": 0,
                "r": 0,
                "b": slider_height,
                "t": 0,
                "pad": 0,
            },  # Add bottom margin for slider
            paper_bgcolor=style["papercolor"],
            plot_bgcolor=style["papercolor"],
            sliders=[
                {
                    "active": 0,
                    "currentvalue": {
                        "font": {"size": 16},
                        "visible": True,
                        "xanchor": "right",
                    },
                    "pad": {"b": 10, "t": 0},  # Adjusted padding
                    "len": 0.9,
                    "x": 0.1,
                    "y": 0,  # Move slider below plot
                    "steps": [],
                }
            ],
        )

        # Create frames for each column
        for i, col_name in enumerate(df.columns):
            series = df[col_name]
            frame_data = [
                bg_hex(),
                data_hex(series),
            ]

            frames.append(go.Frame(data=frame_data, name=str(i)))

            # Add to slider
            slider_steps.append(
                {
                    "args": [
                        [str(i)],
                        {"frame": {"duration": 0, "redraw": True}, "mode": "immediate"},
                    ],
                    "label": col_name,
                    "method": "animate",
                }
            )

            # Set initial display to first column
            if i == 0:
                fig.add_traces(frame_data)

        # Update slider with all steps
        fig.layout.sliders[0].steps = slider_steps
        fig.frames = frames

        # Update axes
        fig.update_xaxes(
            showgrid=False, showticklabels=False, showline=False, visible=False
        )
        fig.update_yaxes(
            showgrid=False, showticklabels=False, showline=False, visible=False
        )

    else:
        # raise error
        raise ValueError("df must be a pd.Series or pd.DataFrame")

    return fig


def looming_stimulus(start_coords, all_coords, n_time=4):
    """
    Generate a list of lists of coordinates for a looming stimulus. The stimulus starts
    at the start_coords and expands outwards in a hexagonal pattern. The stimulus
    expands for n_time steps. Currently the expansion happens one layer at a time.

    Args:
        start_coords (list): List of strings of the form 'x,y' where x and y are the
            coordinates of the starting hexes for the stimulus.
        all_coords (list): List of strings of the form 'x,y' where x and y are the
            coordinates of all hexes in the grid.
        n_time (int): Default=4. Number of time steps for the stimulus to expand.

    Returns:
        stim_str (list): List of lists of strings of the form 'x,y' where x and y are
        the coordinates of the hexes that are stimulated at each time step.
    """
    coords = [tuple(map(float, idx.split(","))) for idx in all_coords]
    x_vals, y_vals = zip(*coords)  # Separate into x and y lists

    # sort and rank x_vals
    x_sorted = sorted(list(set(x_vals)))
    x_to_rank = {x: rank for rank, x in enumerate(x_sorted)}
    rank_to_x = {rank: x for rank, x in enumerate(x_sorted)}
    y_sorted = sorted(list(set(y_vals)))
    y_to_rank = {y: rank for rank, y in enumerate(y_sorted)}
    rank_to_y = {rank: y for rank, y in enumerate(y_sorted)}

    start = [tuple(map(float, idx.split(","))) for idx in start_coords]
    stimulus = []
    stimulus.append(start)
    for atime in range(n_time):
        for x, y in start:
            start_copy = start.copy()
            # hexes above and below x
            if y_to_rank[y] + 2 in rank_to_y:
                start_copy.append((x, rank_to_y[y_to_rank[y] + 2]))
            if y_to_rank[y] - 2 in rank_to_y:
                start_copy.append((x, rank_to_y[y_to_rank[y] - 2]))
            # hexes to the left
            if x_to_rank[x] + 1 in rank_to_x:
                if y_to_rank[y] + 1 in rank_to_y:
                    start_copy.append(
                        (rank_to_x[x_to_rank[x] + 1], rank_to_y[y_to_rank[y] + 1])
                    )
                if y_to_rank[y] - 1 in rank_to_y:
                    start_copy.append(
                        (rank_to_x[x_to_rank[x] + 1], rank_to_y[y_to_rank[y] - 1])
                    )
            # hexes to the right
            if x_to_rank[x] - 1 in rank_to_x:
                if y_to_rank[y] + 1 in rank_to_y:
                    start_copy.append(
                        (rank_to_x[x_to_rank[x] - 1], rank_to_y[y_to_rank[y] + 1])
                    )
                if y_to_rank[y] - 1 in rank_to_y:
                    start_copy.append(
                        (rank_to_x[x_to_rank[x] - 1], rank_to_y[y_to_rank[y] - 1])
                    )

            start = list(set(start_copy))
        stimulus.append(start)

    stim_str = []
    for atime in range(n_time):
        stim_atime = []
        for x, y in stimulus[atime]:
            # Format x and y to remove .0 if they're integers
            x_str = str(int(x)) if x == int(x) else str(x)
            y_str = str(int(y)) if y == int(y) else str(y)
            stim_atime.append(f"{x_str},{y_str}")
        stim_str.append(stim_atime)
    return stim_str


def make_sine_stim(phase=0, amplitude=1, n=8):
    """
    Generate a dictionary of values representing a sine wave stimulus with a given phase
    and amplitude. The sine wave is defined over n points, starting from the given phase.

    Args:
        phase (int): Phase of the sine wave in degrees. Default is 0.
        amplitude (float): Amplitude of the sine wave. Default is 1.
        n (int): Number of points in the sine wave. Default is 8.

    Returns:
        dict: A dictionary where keys are indices from 1 to n, and values are the
        corresponding sine wave values.
    """
    x = (phase % 180) / 180 * np.pi
    x = np.linspace(x, x + np.pi, n)
    y = amplitude * abs(np.sin(x))
    return dict(zip(range(1, n + 1), y))


def plot_mollweide_projection(
    data: pd.Series | pd.DataFrame,
    fig_size: tuple = (900, 700),
    custom_colorscale: Optional[Union[list, str]] = None,
    global_min: Optional[float] = None,
    global_max: Optional[float] = None,
    dataset: str = "Zhao2024",
    marker_size: int = 8,
    value_name: str = "weight",
) -> go.Figure:
    """
    Generates a heatmap to visualize the value of column features per column using the
    mollweide projection.

    Args:
        data (pd.Series | pd.DataFrame): Data with index formatted as strings of the
            form '-12,34', where the first number is the x-coordinate and the second
            number is the y-coordinate. The data to plot. Each column will generate a
            separate frame in the plot.
        fig_size (tuple): Size of the figure in pixels (width, height).
        custom_colorscale (list | str, optional): Custom colorscale for the heatmap. If
            None, defaults to white-to-blue colorscale [[0, "rgb(255, 255, 255)"],
            [1, "rgb(0, 20, 200)"]]. Could also be a string e.g. 'Viridis' or 'Reds'.
        global_min (float | None): Global minimum value for the color scale. If this
            minumum is >0, 0 is used.
        global_max (float | None): Global maximum value for the color scale. If None,
            the maximum value of the data is used.
        dataset (str): The dataset to use for the hexagon locations. Options are:

            - 'Zhao2024': mapping from hexagonal coordinates to 3D coordinates, update from Zhao et al. 2022 (https://www.biorxiv.org/content/10.1101/2022.12.14.520178v1).

        marker_size (int): Size of markers in the plot.

    Returns:
        go.Figure: A Plotly figure object containing the mollweide projection heatmap.
    """

    def cart2sph(xyz: np.array) -> np.array:
        """
        Convert Cartesian to spherical coordinates.
        Theta is polar angle (from +z), phi is angle from +x to +y.
        """
        r = np.sqrt((xyz**2).sum(1))
        theta = np.arccos(xyz[:, 2])
        phi = np.arctan2(xyz[:, 1], xyz[:, 0])
        phi[phi < 0] = phi[phi < 0] + 2 * np.pi
        return np.stack((r, theta, phi), axis=1)

    def sph2Mollweide(thetaphi: np.array) -> np.array:
        """
        Spherical (viewed from outside) to Mollweide,
            cf. https://mathworld.wolfram.com/MollweideProjection.html
        """
        azim = thetaphi[:, 1]
        azim[azim > np.pi] = azim[azim > np.pi] - 2 * np.pi  # longitude/azimuth
        elev = np.pi / 2 - thetaphi[:, 0]  # lattitude/elevation in radian

        N = len(azim)  # number of points
        xy = np.zeros((N, 2))  # output
        for i in range(N):
            theta = np.arcsin(2 * elev[i] / np.pi)
            if np.abs(np.abs(theta) - np.pi / 2) < 0.001:
                xy[i,] = [
                    2 * np.sqrt(2) / np.pi * azim[i] * np.cos(theta),
                    np.sqrt(2) * np.sin(theta),
                ]
            else:
                # to calculate theta
                dtheta = 1
                while dtheta > 1e-3:
                    theta_new = theta - (
                        2 * theta + np.sin(2 * theta) - np.pi * np.sin(elev[i])
                    ) / (2 + 2 * np.cos(2 * theta))
                    dtheta = np.abs(theta_new - theta)
                    theta = theta_new
                xy[i,] = [
                    2 * np.sqrt(2) / np.pi * azim[i] * np.cos(theta),
                    np.sqrt(2) * np.sin(theta),
                ]
        return xy

    def create_mollweide_guidelines():
        """
        Create Mollweide projection guidelines as plotly traces
        """
        traces = []

        # Create meridians
        ww = np.stack((np.linspace(0, 180, 19), np.repeat(-180, 19)), axis=1)
        w = np.stack((np.linspace(180, 0, 19), np.repeat(-90, 19)), axis=1)
        m = np.stack((np.linspace(0, 180, 19), np.repeat(0, 19)), axis=1)
        e = np.stack((np.linspace(180, 0, 19), np.repeat(90, 19)), axis=1)
        ee = np.stack((np.linspace(0, 180, 19), np.repeat(180, 19)), axis=1)
        pts = np.vstack((ww, w, m, e, ee))
        rtp = np.insert(pts / 180 * np.pi, 0, np.repeat(1, pts.shape[0]), axis=1)
        meridians_xy = sph2Mollweide(rtp[:, 1:3])

        traces.append(
            go.Scatter(
                x=meridians_xy[:, 0],
                y=meridians_xy[:, 1],
                mode="lines",
                line=dict(color="lightgrey", width=0.5),
                showlegend=False,
                hoverinfo="skip",
            )
        )

        # Create parallels
        for lat in [45, 90, 135]:
            pts = np.stack((np.repeat(lat, 37), np.linspace(-180, 180, 37)), axis=1)
            rtp = np.insert(pts / 180 * np.pi, 0, np.repeat(1, pts.shape[0]), axis=1)
            parallel_xy = sph2Mollweide(rtp[:, 1:3])

            traces.append(
                go.Scatter(
                    x=parallel_xy[:, 0],
                    y=parallel_xy[:, 1],
                    mode="lines",
                    line=dict(color="lightgrey", width=0.5),
                    showlegend=False,
                    hoverinfo="skip",
                )
            )

        return traces

    def create_data_scatter(series_data, x_coords, y_coords, column_name=None):
        """Create scatter plot for data points"""
        return go.Scatter(
            x=x_coords,
            y=y_coords,
            mode="markers",
            marker=dict(
                color=series_data.values,
                colorscale=custom_colorscale,
                cmin=global_min,
                cmax=global_max,
                size=marker_size,
                colorbar=dict(
                    title=dict(text=value_name, side="right"),
                ),
            ),
            customdata=np.stack([x_coords, y_coords, series_data.values], axis=-1),
            hovertemplate="x: %{customdata[0]:.2f}<br>y: %{customdata[1]:.2f}<br>%{text}: %{customdata[2]:.4f}<extra></extra>",
            text=[value_name] * len(series_data),
            showlegend=False,
        )

    # Default colorscale
    if custom_colorscale is None:
        custom_colorscale = [[0, "rgb(255, 255, 255)"], [1, "rgb(0, 20, 200)"]]

    # Clean data - remove NaN indices
    data = data[(data.index != "nan") & (~data.index.isnull())]

    # Convert string indices to coordinate arrays
    coords = [tuple(map(float, idx.split(","))) for idx in data.index]
    coord_array = np.array(coords)

    # Get global min and max for consistent color scale
    vals = data.to_numpy() if isinstance(data, pd.DataFrame) else data.values
    if global_min is None:
        global_min = min(0, vals.min())
    if global_max is None:
        global_max = vals.max()

    # Load eyemap data and convert coordinates
    ucl_hex = load_dataset(dataset)
    rtp2 = cart2sph(ucl_hex[["x", "y", "z"]].values)
    xy = sph2Mollweide(rtp2[:, 1:3])
    xy[:, 0] = -xy[:, 0]  # flip x axis
    xypq_moll = np.concatenate((xy, ucl_hex[["p", "q"]].values), axis=1)
    xypq_moll = pd.DataFrame(xypq_moll, columns=["x", "y", "p", "q"])
    xypq_moll[["p", "q"]] = xypq_moll[["p", "q"]].astype(int)

    # Convert data coordinates to Mollweide
    hex1_id = (coord_array[:, 1] - coord_array[:, 0]) / 2
    hex2_id = (coord_array[:, 1] + coord_array[:, 0]) / 2

    coord_df = pd.DataFrame({"hex1_id": hex1_id, "hex2_id": hex2_id}, index=data.index)

    merged_coords = coord_df.merge(
        xypq_moll, left_on=["hex1_id", "hex2_id"], right_on=["q", "p"], how="left"
    )

    x_mollweide = merged_coords["x"].values
    y_mollweide = merged_coords["y"].values

    # Create figure
    fig = go.Figure()

    # Add guidelines
    guidelines = create_mollweide_guidelines()
    for trace in guidelines:
        fig.add_trace(trace)

    # Handle single series vs DataFrame
    if isinstance(data, pd.Series) or (
        isinstance(data, pd.DataFrame) and len(data.columns) == 1
    ):
        if isinstance(data, pd.DataFrame):
            data = data.iloc[:, 0]

        # Single plot
        fig.add_trace(create_data_scatter(data, x_mollweide, y_mollweide))

    elif isinstance(data, pd.DataFrame):
        # Multiple columns - create frames for slider
        frames = []
        slider_steps = []

        # Create frames for each column
        for i, col_name in enumerate(data.columns):
            series = data[col_name]

            # Create frame data (guidelines + data scatter)
            frame_traces = guidelines + [
                create_data_scatter(series, x_mollweide, y_mollweide)
            ]
            frames.append(go.Frame(data=frame_traces, name=str(i)))

            # Add slider step
            slider_steps.append(
                {
                    "args": [
                        [str(i)],
                        {"frame": {"duration": 0, "redraw": True}, "mode": "immediate"},
                    ],
                    "label": col_name,
                    "method": "animate",
                }
            )

            # Set initial display to first column
            if i == 0:
                fig.add_trace(
                    create_data_scatter(series, x_mollweide, y_mollweide, col_name)
                )

        # Add frames to figure
        fig.frames = frames

        # Add slider
        fig.update_layout(
            sliders=[
                {
                    "active": 0,
                    "currentvalue": {
                        "font": {"size": 16},
                        "visible": True,
                        "xanchor": "right",
                    },
                    "pad": {"b": 10, "t": 50},
                    "len": 0.9,
                    "x": 0.1,
                    "y": 0,
                    "steps": slider_steps,
                }
            ]
        )

    # Update layout
    fig.update_layout(
        width=fig_size[0],
        height=fig_size[1],
        xaxis=dict(
            range=[-np.pi, np.pi],
            scaleanchor="y",
            scaleratio=1,
            showgrid=False,
            showticklabels=False,
            showline=False,
            visible=False,
        ),
        yaxis=dict(
            range=[-np.pi / 2, np.pi / 2],
            showgrid=False,
            showticklabels=False,
            showline=False,
            visible=False,
        ),
        plot_bgcolor="white",
        paper_bgcolor="white",
        margin=dict(
            l=0,
            r=0,
            t=50,
            b=50 if isinstance(data, pd.DataFrame) and len(data.columns) > 1 else 0,
        ),
    )

    return fig



import unittest

import numpy as np
import pandas as pd
import torch
from scipy.sparse import csr_matrix
import torch

from connectome_interpreter.activation_maximisation import (
    MultilayeredNetwork,
    TargetActivation,
    activation_maximisation,
    activations_to_df,
    activations_to_df_batched,
    get_neuron_activation,
)


class TestMultilayeredNetwork(unittest.TestCase):
    def setUp(self):
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.num_neurons = 10
        self.num_sensory = 4
        self.num_layers = 3
        self.batch_size = 2

        # Create a dense matrix and convert it to a scipy sparse matrix
        dense_weights = np.random.rand(self.num_neurons, self.num_neurons)
        dense_weights = dense_weights / dense_weights.sum(axis=1, keepdims=True)
        dense_weights[:, :3] = -dense_weights[:, :3]
        self.all_weights = csr_matrix(dense_weights)  # Convert to scipy sparse matrix
        self.sensory_indices = list(range(self.num_sensory))

        self.model = MultilayeredNetwork(
            self.all_weights, self.sensory_indices, num_layers=self.num_layers
        ).to(self.device)

    def test_initialization(self):
        # self.assertEqual(self.model.num_layers, self.num_layers)
        # self.assertEqual(len(self.model.sensory_indices), self.num_sensory)
        # self.assertTrue(
        #     torch.equal(
        #         self.model.all_weights.to_dense(),
        #         torch.tensor(self.all_weights.toarray(), device=self.device),
        #     )
        # )
        self.assertEqual(self.model.num_layers, self.num_layers)
        self.assertEqual(len(self.model.sensory_indices), self.num_sensory)

        # Convert both to numpy arrays for comparison
        model_weights = self.model.all_weights.to_dense().cpu().numpy()
        expected_weights = self.all_weights.toarray()

        # Use numpy's allclose for a more tolerant comparison
        self.assertTrue(
            np.allclose(model_weights, expected_weights, rtol=1e-5, atol=1e-5),
            "Weights matrices are not equal within tolerance",
        )

    def test_forward_pass_2d(self):
        print("testing forward pass 2d")
        input_tensor = torch.rand(self.num_sensory, self.num_layers).to(self.device)
        output = self.model(input_tensor)

        expected_shape = (self.num_neurons, self.num_layers)
        self.assertEqual(output.shape, expected_shape)
        self.assertTrue(torch.all(output >= -1) and torch.all(output <= 1))

    def test_forward_pass_3d(self):
        print("testing forward pass 3d")
        input_tensor = torch.rand(
            self.batch_size, self.num_sensory, self.num_layers
        ).to(self.device)
        output = self.model(input_tensor)

        expected_shape = (self.batch_size, self.num_neurons, self.num_layers)
        self.assertEqual(output.shape, expected_shape)
        self.assertTrue(torch.all(output >= -1) and torch.all(output <= 1))


class TestTargetActivation(unittest.TestCase):
    def setUp(self):
        self.dict_targets = {0: {1: 0.5, 2: 0.8}, 1: {0: 0.3}}
        self.df_targets = pd.DataFrame(
            [
                {"batch": 0, "layer": 0, "neuron": 1, "value": 0.5},
                {"batch": 0, "layer": 0, "neuron": 2, "value": 0.8},
                {"batch": 1, "layer": 1, "neuron": 0, "value": 0.3},
            ]
        )

    def test_dict_initialization(self):
        target = TargetActivation(targets=self.dict_targets, batch_size=2)
        self.assertEqual(target.batch_size, 2)

        batch_targets = target.get_batch_targets(0)
        self.assertEqual(batch_targets[0][1], 0.5)

    def test_df_initialization(self):
        target = TargetActivation(targets=self.df_targets)
        self.assertEqual(target.batch_size, 2)

        batch_targets = target.get_batch_targets(1)
        self.assertEqual(batch_targets[1][0], 0.3)


class TestActivationMaximisation(unittest.TestCase):
    def setUp(self):
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

        # Create a dense matrix and convert it to a scipy sparse matrix
        dense_weights = np.random.rand(10, 10)
        dense_weights = dense_weights / dense_weights.sum(axis=1, keepdims=True)
        dense_weights[:, :3] = -dense_weights[:, :3]
        self.all_weights = csr_matrix(dense_weights)  # Convert to scipy sparse matrix

        self.model = MultilayeredNetwork(
            self.all_weights,
            sensory_indices=[0, 1, 2, 3],
            num_layers=3,
        ).to(self.device)

        self.targets = TargetActivation(
            {0: {0: 0.5, 1: 0.8}, 1: {2: 0.3}}, batch_size=2
        )

    def test_basic_optimization(self):
        print("testing basic optimization")

        result = activation_maximisation(
            self.model,
            self.targets,
            num_iterations=10,
            in_reg_lambda=1e-3,
            out_reg_lambda=1e-3,
            wandb=False,
            device=self.device,
        )

        input_tensor, output, act_losses, *_ = result
        expected_shape = (2, 4, 3)  # (batch_size, num_sensory, num_layers)
        self.assertEqual(input_tensor.shape, expected_shape)
        self.assertTrue(act_losses[-1] <= act_losses[0])

    def test_custom_regularization(self):
        print("testing custom regularization")

        custom_reg = {
            "in": lambda x: torch.sum(torch.abs(x)),
            "out": lambda x: torch.sum(torch.abs(x)),
        }

        result = activation_maximisation(
            self.model,
            self.targets,
            custom_reg_functions=custom_reg,
            num_iterations=10,
            wandb=False,
            device=self.device,
        )

        _, _, _, out_reg_losses, in_reg_losses, _ = result
        self.assertTrue(len(in_reg_losses) > 0 and len(out_reg_losses) > 0)


class TestActivationsToDF(unittest.TestCase):
    def setUp(self):
        self.weights = np.array([[0.5, 0.3, 0.0], [0.0, 0.4, 0.2], [0.0, 0.0, 0.6]])
        self.input_act = np.array([[0.8, 0.6], [0.7, 0.5]])
        self.output_act = np.array([[0.8, 0.6], [0.7, 0.5], [0.6, 0.4]])
        self.sensory_indices = [0, 1]

    def test_basic_functionality(self):
        paths = activations_to_df(
            self.weights, self.input_act, self.output_act, self.sensory_indices
        )

        expected_columns = [
            "pre",
            "post",
            "weight",
            "layer",
            "pre_activation",
            "post_activation",
        ]
        self.assertTrue(all(col in paths.columns for col in expected_columns))
        self.assertEqual(paths["layer"].nunique(), self.output_act.shape[1])

    def test_sparse_input(self):
        sparse_weights = csr_matrix(self.weights)

        paths_dense = activations_to_df(
            self.weights, self.input_act, self.output_act, self.sensory_indices
        )
        paths_sparse = activations_to_df(
            sparse_weights,
            self.input_act,
            self.output_act,
            self.sensory_indices,
        )

        pd.testing.assert_frame_equal(paths_dense, paths_sparse)


class TestActivationsToDFBatched(unittest.TestCase):
    def setUp(self):
        self.weights = np.array([[0.5, 0.3, 0.0], [0.0, 0.4, 0.2], [0.0, 0.0, 0.6]])
        self.batched_input = np.array(
            [[[0.8, 0.6], [0.7, 0.5]], [[0.6, 0.4], [0.5, 0.3]]]
        )
        self.batched_output = np.array(
            [
                [[0.8, 0.6], [0.7, 0.5], [0.6, 0.4]],
                [[0.6, 0.4], [0.5, 0.3], [0.4, 0.2]],
            ]
        )
        self.sensory_indices = [0, 1]

    def test_batched_processing(self):
        print("testing batched proccessing")
        paths = activations_to_df_batched(
            self.weights,
            self.batched_input,
            self.batched_output,
            self.sensory_indices,
        )

        self.assertTrue("batch" in paths.columns)
        self.assertEqual(paths["batch"].nunique(), self.batched_input.shape[0])


class TestGetNeuronActivation(unittest.TestCase):

    def test_2d_output_with_groups(self):
        output = torch.tensor([[0.1, 0.2, 0.3], [0.4, 0.5, 0.6], [0.7, 0.8, 0.9]])
        neuron_indices = [0, 2]
        idx_to_group = {0: "A", 2: "B"}

        df = get_neuron_activation(output, neuron_indices, idx_to_group=idx_to_group)

        expected_df = pd.DataFrame(
            {
                "group": ["A", "B"],
                "time_0": [0.1, 0.7],
                "time_1": [0.2, 0.8],
                "time_2": [0.3, 0.9],
            }
        )
        pd.testing.assert_frame_equal(df, expected_df, check_dtype=False)

    def test_2d_output_without_groups(self):
        output = torch.tensor([[0.1, 0.2, 0.3], [0.4, 0.5, 0.6], [0.7, 0.8, 0.9]])
        neuron_indices = [1]

        df = get_neuron_activation(output, neuron_indices)

        expected_df = pd.DataFrame(
            {"group": [1], "time_0": [0.4], "time_1": [0.5], "time_2": [0.6]}
        )
        pd.testing.assert_frame_equal(df, expected_df, check_dtype=False)

    def test_3d_output_with_batch_names_and_groups(self):
        output = torch.tensor([[[0.1, 0.2], [0.3, 0.4]], [[0.5, 0.6], [0.7, 0.8]]])
        neuron_indices = [0]
        batch_names = ["batch_1", "batch_2"]
        idx_to_group = {0: "A"}

        df = get_neuron_activation(
            output,
            neuron_indices,
            batch_names=batch_names,
            idx_to_group=idx_to_group,
        )

        expected_df = pd.DataFrame(
            {
                "batch_name": ["batch_1", "batch_2"],
                "group": ["A", "A"],
                "time_0": [0.1, 0.5],
                "time_1": [0.2, 0.6],
            }
        )
        pd.testing.assert_frame_equal(df, expected_df, check_dtype=False)

    def test_3d_output_without_batch_names(self):
        output = torch.tensor([[[0.1, 0.2], [0.3, 0.4]], [[0.5, 0.6], [0.7, 0.8]]])
        neuron_indices = [1]

        df = get_neuron_activation(output, neuron_indices)

        expected_df = pd.DataFrame(
            {
                "batch_name": ["batch_0", "batch_1"],
                "group": [1, 1],
                "time_0": [0.3, 0.7],
                "time_1": [0.4, 0.8],
            }
        )
        pd.testing.assert_frame_equal(df, expected_df, check_dtype=False)

    def test_batch_names_mismatch(self):
        output = torch.tensor([[[0.1, 0.2], [0.3, 0.4]], [[0.5, 0.6], [0.7, 0.8]]])
        neuron_indices = [1]
        batch_names = ["batch_1"]

        with self.assertRaises(ValueError):
            get_neuron_activation(output, neuron_indices, batch_names=batch_names)


# Add these test classes to your existing test file


class TestMultilayeredNetworkEnhanced(unittest.TestCase):
    def setUp(self):
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.num_neurons = 10
        self.num_sensory = 4
        self.num_layers = 3

        # Create weights and mappings
        dense_weights = np.random.rand(self.num_neurons, self.num_neurons)
        dense_weights = dense_weights / dense_weights.sum(axis=1, keepdims=True)
        self.all_weights = csr_matrix(dense_weights)
        self.sensory_indices = list(range(self.num_sensory))

        # Create idx_to_group mapping
        self.idx_to_group = {i: f"type_{i//3}" for i in range(self.num_neurons)}

    def test_trainable_parameters_initialization(self):
        """Test initialization with trainable parameters"""
        model = MultilayeredNetwork(
            self.all_weights,
            self.sensory_indices,
            idx_to_group=self.idx_to_group,
            default_bias=0.2,
            tanh_steepness=3.0,
        ).to(self.device)

        # Check that parameters exist
        self.assertIsNotNone(model.slope)
        self.assertIsNotNone(model.biases)
        self.assertIsNotNone(model.indices)

        # Check parameter shapes
        num_types = len(set(self.idx_to_group.values()))
        self.assertEqual(model.slope.shape[0], num_types)
        self.assertEqual(model.biases.shape[0], num_types)

    def test_dict_parameter_values(self):
        """Test initialization with dictionary parameter values"""
        bias_dict = {"type_0": 0.1, "type_1": 0.2, "type_2": 0.3}
        slope_dict = {"type_0": 2.0, "type_1": 4.0, "type_2": 6.0}

        model = MultilayeredNetwork(
            self.all_weights,
            self.sensory_indices,
            idx_to_group=self.idx_to_group,
            bias_dict=bias_dict,
            slope_dict=slope_dict,
        ).to(self.device)

        # Check that parameters were set correctly
        # The order depends on set() ordering, so check individual values
        unique_types = sorted(set(self.idx_to_group.values()))
        for i, type_name in enumerate(unique_types):
            if type_name in bias_dict:
                expected_bias = bias_dict[type_name]
            else:
                expected_bias = 0  # default
            self.assertAlmostEqual(model.raw_biases[i].item(), expected_bias, places=6)

    def test_custom_activation_function(self):
        """Test custom activation function"""

        def custom_activation(x):
            return torch.sigmoid(x)

        model = MultilayeredNetwork(
            self.all_weights,
            self.sensory_indices,
            activation_function=custom_activation,
        ).to(self.device)

        input_tensor = torch.rand(self.num_sensory, self.num_layers).to(self.device)
        output = model(input_tensor)

        # Check that output is in sigmoid range [0, 1]
        self.assertTrue(torch.all(output >= 0) and torch.all(output <= 1))

    def test_backward_compatibility(self):
        """Test that model works without new parameters (backward compatibility)"""
        model = MultilayeredNetwork(
            self.all_weights, self.sensory_indices, num_layers=self.num_layers
        ).to(self.device)

        # Check that trainable parameters are None
        self.assertIsNone(model.slope)
        self.assertIsNone(model.biases)
        self.assertIsNone(model.indices)

        # Check that forward pass still works
        input_tensor = torch.rand(self.num_sensory, self.num_layers).to(self.device)
        output = model(input_tensor)
        self.assertEqual(output.shape, (self.num_neurons, self.num_layers))

    def test_parameter_gradient_control(self):
        """Test set_param_grads method"""
        model = MultilayeredNetwork(
            self.all_weights, self.sensory_indices, idx_to_group=self.idx_to_group
        ).to(self.device)

        # Initially parameters shouldn't require grad
        self.assertFalse(model.slope.requires_grad)
        self.assertFalse(model.raw_biases.requires_grad)

        # Enable gradients
        model.set_param_grads(slopes=True, raw_biases=True)
        self.assertTrue(model.slope.requires_grad)
        self.assertTrue(model.raw_biases.requires_grad)

        # Disable gradients
        model.set_param_grads(slopes=False, raw_biases=False)
        self.assertFalse(model.slope.requires_grad)
        self.assertFalse(model.raw_biases.requires_grad)


class TestContextManagers(unittest.TestCase):
    def setUp(self):
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        dense_weights = np.random.rand(5, 5)
        self.all_weights = csr_matrix(dense_weights)
        self.sensory_indices = [0, 1]
        self.idx_to_group = {i: f"type_{i}" for i in range(5)}

    def test_training_mode_context(self):
        """Test training_mode context manager"""
        from connectome_interpreter.activation_maximisation import training_mode

        model = MultilayeredNetwork(
            self.all_weights, self.sensory_indices, idx_to_group=self.idx_to_group
        ).to(self.device)

        # Initially no gradients
        self.assertFalse(model.slope.requires_grad)
        self.assertFalse(model.raw_biases.requires_grad)

        # Inside context, gradients should be enabled
        with training_mode(model):
            self.assertTrue(model.slope.requires_grad)
            self.assertTrue(model.raw_biases.requires_grad)

        # After context, gradients should be disabled
        self.assertFalse(model.slope.requires_grad)
        self.assertFalse(model.raw_biases.requires_grad)


class TestTrainModelEnhanced(unittest.TestCase):
    def setUp(self):
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        dense_weights = np.random.rand(6, 6) * 0.1
        self.all_weights = csr_matrix(dense_weights)
        self.sensory_indices = [0, 1]
        self.idx_to_group = {i: f"type_{i//2}" for i in range(6)}

        self.model = MultilayeredNetwork(
            self.all_weights,
            self.sensory_indices,
            num_layers=2,
            idx_to_group=self.idx_to_group,
        ).to(self.device)

    def test_time_series_targets(self):
        """Test training with time series targets (layer column)"""
        from connectome_interpreter.activation_maximisation import train_model

        # Create inputs and time series targets
        inputs = torch.rand(4, 2, 2).to(self.device)
        targets = pd.DataFrame(
            [
                {"batch": 0, "neuron_idx": 2, "layer": 0, "value": 0.5},
                {"batch": 0, "neuron_idx": 2, "layer": 1, "value": 0.8},
                {"batch": 1, "neuron_idx": 3, "layer": 0, "value": 0.3},
                {"batch": 2, "neuron_idx": 4, "layer": 1, "value": 0.7},
                {"batch": 3, "neuron_idx": 5, "layer": 0, "value": 0.4},
            ]
        )

        # Train model
        result = train_model(self.model, inputs, targets, num_epochs=5, wandb=False)

        model, history, *_ = result

        # Check that training occurred
        self.assertTrue(len(history["loss"]) > 0)
        self.assertIsInstance(history["loss"][0], float)

    def test_backward_compatible_targets(self):
        """Test training with old format targets (no layer column)"""
        from connectome_interpreter.activation_maximisation import train_model

        inputs = torch.rand(4, 2, 2).to(self.device)
        targets = pd.DataFrame(
            [
                {"batch": 0, "neuron_idx": 2, "value": 0.5},
                {"batch": 1, "neuron_idx": 3, "value": 0.3},
                {"batch": 2, "neuron_idx": 4, "value": 0.7},
                {"batch": 3, "neuron_idx": 5, "value": 0.4},
            ]
        )

        # Should work without error
        result = train_model(self.model, inputs, targets, num_epochs=3, wandb=False)

        model, history, *_ = result
        self.assertTrue(len(history["loss"]) > 0)


class TestSaliencyEnhanced(unittest.TestCase):
    def setUp(self):
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        dense_weights = np.random.rand(8, 8) * 0.1
        self.all_weights = csr_matrix(dense_weights)
        self.sensory_indices = [0, 1, 2]
        self.idx_to_group = {i: f"type_{i//2}" for i in range(8)}

    def test_saliency_with_trainable_model(self):
        """Test saliency computation with trainable parameters"""
        from connectome_interpreter.activation_maximisation import saliency

        model = MultilayeredNetwork(
            self.all_weights,
            self.sensory_indices,
            num_layers=2,
            idx_to_group=self.idx_to_group,
        ).to(self.device)

        input_tensor = torch.rand(3, 2).to(self.device)
        neurons_of_interest = {0: [4, 5], 1: [6, 7]}

        saliency_maps = saliency(
            model, input_tensor, neurons_of_interest, device=self.device
        )

        # Check output shape and that gradients were computed
        self.assertEqual(saliency_maps.shape, input_tensor.shape)
        self.assertFalse(torch.allclose(saliency_maps, torch.zeros_like(saliency_maps)))

    def test_saliency_methods(self):
        """Test different saliency methods"""
        from connectome_interpreter.activation_maximisation import saliency

        model = MultilayeredNetwork(
            self.all_weights, self.sensory_indices, num_layers=2
        ).to(self.device)

        input_tensor = torch.rand(3, 2).to(self.device)
        neurons_of_interest = {0: [4], 1: [5]}

        # Test vanilla saliency
        vanilla_sal = saliency(
            model,
            input_tensor,
            neurons_of_interest,
            method="vanilla",
            device=self.device,
        )

        # Test input_x_gradient saliency
        ixg_sal = saliency(
            model,
            input_tensor,
            neurons_of_interest,
            method="input_x_gradient",
            device=self.device,
        )

        # Should produce different results
        self.assertFalse(torch.allclose(vanilla_sal, ixg_sal))

    def test_saliency_invalid_method(self):
        """Test saliency with invalid method"""
        from connectome_interpreter.activation_maximisation import saliency

        model = MultilayeredNetwork(
            self.all_weights, self.sensory_indices, num_layers=2
        ).to(self.device)

        input_tensor = torch.rand(3, 2).to(self.device)
        neurons_of_interest = {0: [4]}

        with self.assertRaises(ValueError):
            saliency(
                model,
                input_tensor,
                neurons_of_interest,
                method="invalid_method",
                device=self.device,
            )


class TestTargetActivationEnhanced(unittest.TestCase):
    def test_time_series_targets_dict(self):
        """Test TargetActivation with time series targets from dict"""
        targets_dict = {0: {1: 0.5, 2: 0.8}, 1: {0: 0.3, 3: 0.7}}  # layer 0  # layer 1
        target = TargetActivation(targets=targets_dict, batch_size=3)

        # Check that all batches have the same targets
        batch_0_targets = target.get_batch_targets(0)
        batch_2_targets = target.get_batch_targets(2)
        self.assertEqual(batch_0_targets, batch_2_targets)

        # Check layer structure
        self.assertIn(0, batch_0_targets)  # layer 0
        self.assertIn(1, batch_0_targets)  # layer 1


class TestDivisiveNormalization(unittest.TestCase):
    def setUp(self):
        # Simple 2x2 weight matrix for tests
        dense = np.array([[0.0, 0.0], [0.0, 0.0]], dtype=float)
        self.weights = csr_matrix(dense)
        self.sensory_indices = [0]
        self.idx_to_group = {0: "A", 1: "B"}

    def test_missing_idx_to_group_raises(self):
        dn = {"A": ["B"]}
        with self.assertRaises(AssertionError):
            MultilayeredNetwork(
                self.weights, self.sensory_indices, divisive_normalization=dn
            )

    def test_positive_divnorm_weight_raises(self):
        dense = np.array([[0.0, 0.0], [1.0, 0.0]], dtype=float)  # weight[1,0] positive
        weights = csr_matrix(dense)
        dn = {"A": ["B"]}
        with self.assertRaises(ValueError):
            MultilayeredNetwork(
                weights,
                self.sensory_indices,
                idx_to_group=self.idx_to_group,
                divisive_normalization=dn,
            )

    def test_weight_removal_and_storage(self):
        dense = np.array([[0.0, -0.5], [0.3, 0.0]], dtype=float)  # from B to A, -0.5
        weights = csr_matrix(dense)
        dn = {"B": ["A"]}
        model = MultilayeredNetwork(
            weights,
            self.sensory_indices,
            idx_to_group=self.idx_to_group,
            divisive_normalization=dn,
        )
        dense_out = model.all_weights.to_dense().cpu().numpy()
        # entry (post=0, pre=1) should be removed -> zero
        self.assertEqual(dense_out[0, 1], 0.0)
        # stored divnorm_weights should contain original -0.5
        self.assertTrue(
            torch.allclose(
                model.divnorm_weights,
                torch.tensor([-0.5], dtype=torch.float32).to(
                    model.divnorm_weights.device
                ),
                atol=1e-6,
            )
        )
        # stored indices should point to (0, 1) -> pre=1, post=0
        idx = model.divnorm_indices.cpu().numpy()
        self.assertTrue(np.array_equal(idx, np.array([[0], [1]])))

    def test_divisive_strength_param(self):
        dense = np.array([[0.0, -0.2], [0.0, 0.0]], dtype=float)
        weights = csr_matrix(dense)
        dn = {"B": ["A"]}  # one pres type
        # custom strength
        ds = {"B": 0.3}
        model = MultilayeredNetwork(
            weights,
            self.sensory_indices,
            idx_to_group=self.idx_to_group,
            divisive_normalization=dn,
            divisive_strength=ds,
        )
        # one strength value
        self.assertEqual(model.divisive_strength.numel(), 1)
        self.assertFalse(model.divisive_strength.requires_grad)
        self.assertAlmostEqual(model.divisive_strength.item(), 0.3, places=6)

    def test_divnorm_identity_when_none(self):
        dense = np.array([[0.1, 0.2], [0.3, 0.4]], dtype=float)
        weights = csr_matrix(dense)
        model = MultilayeredNetwork(weights, self.sensory_indices)
        slopes = torch.tensor([1.0, 2.0])
        x_prev = torch.tensor([0.5, 0.5])
        out = model.divnorm(slopes.clone(), x_prev)
        self.assertTrue(torch.allclose(out, slopes))

    def test_divnorm_effect_forward(self):
        # If pre-activation=1, weight=-1 and divisive_strength=1, new slope = 0 -> no activation
        dense = np.array(
            [
                [0.0, 0.0, 0.0],
                [-0.5, 0.0, 1.0],  # C excites B, A inhibits B
                [0.0, 0.0, 0.0],
            ],
            dtype=float,
        )
        weights = csr_matrix(dense)
        dn = {"A": ["B"]}
        ds = {"A": 20}  # a very big divisive strength, should set slope to 0
        model = MultilayeredNetwork(
            weights,
            sensory_indices=[0, 2],
            tanh_steepness=5,
            idx_to_group={0: "A", 1: "B", 2: "C"},
            divisive_normalization=dn,
            divisive_strength=ds,
        )
        # one layer, one sensory neuron => shape (sensory, layers) = (2,2)
        inp = torch.tensor([[1.0, 0.0], [1.0, 0.0]], dtype=torch.float32)
        out = model(inp)
        # post neuron idx=1 should receive zero activation
        self.assertAlmostEqual(out[1, 0].item(), 0.0, places=6)

        # and on the other hand, without divisive normalization, it should > 0
        model_no_dn = MultilayeredNetwork(
            weights, sensory_indices=[0, 2], idx_to_group={0: "A", 1: "B", 2: "C"}
        )
        out_no_dn = model_no_dn(inp)
        self.assertGreater(out_no_dn[1, 0].item(), 0.0)
